/** 
 *   5850 Absolute shaft encoder driver.
 *
 *   (c) Copyright 2003 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: 
 *      *) spin_lock.
 *      -) data communication.
 *      *) if load fail: exit is not clean.
 *      -) file_operations: read 
 *      -) file_operations: ioctl
 *      *) method for get position on sid. clock fail condition.
 *      *)
 *      *)
 *      *)
 *
 * 06/09/2004: Remove ATOMIC_VAL conditional flag.
 *             Add low pas filter for anemometer (circular buffer).
 **/

#define E5850_VERSION "0.0.3"

#define USING_PROC_FS

#include <linux/module.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/delay.h>  /* udelay */
#include "e5850.h"

#ifndef CONFIG_PROC_FS
#undef USING_PROC_FS
#endif

#define OUR_NAME "encoder5850"

/* Port definitions: */
#define BASE_IO 0x378

/**
 * The pinout of the parallel port for absolute encoder 5850 and
 * anemometer
 *
 *      Pin  |  Signal                | Bit  |  Port
 *      -----+------------------------+------+-----------------
 *        2  |  Pulse for 5850's      |  0   |  Data   (base+0)
 *        3  |  Pulse for Anemometer  |  1   |  Data   (base+0)
 *        4  |  Clockwise encoder 1   |  2   |  Data   (base+0)
 *        5  |  Set 0 for encoder 1   |  3   |  Data   (base+0)
 *        6  |  Clockwise encoder 0   |  4   |  Data   (base+0)
 *        7  |  Set 0 for encoder 0   |  5   |  Data   (base+0)
 *        8  |  Anem. AD enable/desab.|  6   |  Data   (base+0)
 *       10  |  Sidereal clock (IRQ)  |  6   |  Status (base+1)
 *       12  |  Dara read encoder 1   |  5   |  Status (base+1)
 *       13  |  Data read encoder 0   |  4   |  Status (base+1)
 *       15  |  Data read anemometer 0|  3   |  Status (base+1)
 *
 *  Control Port (base+2) only use bit 4 for enabling interrupts.
 **/

/* Pulse use bit 0 of data port (pin 2) for encoders and bit 1 (pin 3)
   for anemometer */
#define PULSE_UP   (0x03 | data_current) /* 1|2 */
#define PULSE_DOWN (0x00 | data_current)

/* first encoder bits positions (pin 13) */
#define MASK_0 0x10 
#define OFFSET_0 4
#define CCLOCKWISE_0 0x10
#define SETZ_0 0x20
/* second encoder bits positions (pin 12) */
#define MASK_1 0x20 
#define OFFSET_1 5
#define CCLOCKWISE_1 0x04
#define SETZ_1 0x08
/* anemometer bits positions (pin 15) */
#define MASK_2 0x08 
#define OFFSET_2 3
#define AAD_DESABLE 0x40 /* Desable Anemometer AD */

/* anemometer buffer (low pas filter ~ 2s ) */
#define ANEMOMETER_BUFF_LEN 8
#define ANEMOMETER_BUFF_SHIFT 3

#define ENABLE_INTERRUPT()   (outb(0x10, e5850_base+2))
#define DISABLE_INTERRUPT() (outb(0x00, e5850_base + 2))

static int major = 0; /* dynamic by default */
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major,"Major number for /dev/e5850; dynamic by default.");

/* default is the first printer port on PC's. "e5850_base" is there
   too because it's what we want to use in the code */
static unsigned long base = BASE_IO;
unsigned long e5850_base = 0;
MODULE_PARM(base, "i");
MODULE_PARM_DESC(base,"Base io port (default=0x378)");

/* The interrupt line is undefined by default. "e5850_irq" is as above */
static int irq = -1;
volatile int e5850_irq = -1;
MODULE_PARM(irq, "i");
MODULE_PARM_DESC(irq,"irq line");

static int nrbits = E5850_NRBITS;
volatile int e5850_nrb;
MODULE_PARM(nrbits, "i");
MODULE_PARM_DESC(nrbits,"Number of bits resolution");


atomic_t test_irq;
#ifdef USING_PROC_FS
struct proc_dir_entry *e5850_proc;
#endif
atomic_t encoder_0;
atomic_t encoder_1;
atomic_t anemometer_sum;

/* current value in data port: */
static int data_current;

static unsigned long e5850_is_open;
static unsigned long e5850_is_running;
static unsigned long e5850_fail_clock;

/* definitions for cicular bufer */
static unsigned long e5850_buffer = 0;
volatile unsigned long e5850_head;
volatile unsigned long e5850_tail;
DECLARE_WAIT_QUEUE_HEAD(e5850_queue);
/* for anemometer: */
static int anemometer_buff[ANEMOMETER_BUFF_LEN];


static inline void anemometer_push(int val)
{
	static int anemometer_index = 0;
	int inc;

	/* In atomic_t type, the full integer range may not be
	   available; thus, you should not count on an atomic_t
	   holding more than 24 bits (0 - 16777215).  */

	//atomic_sub(anemometer_buff[anemometer_index], &anemometer_sum);
	//atomic_add(val, &anemometer_sum);
	inc = val - anemometer_buff[anemometer_index];
	if ( inc > 0) {
		atomic_add(inc, &anemometer_sum);
	}
	else {
		/* clean the sign */ 
		atomic_sub( -inc , &anemometer_sum);
	}
	anemometer_buff[anemometer_index] = val;
	anemometer_index++;
	if ( anemometer_index == ANEMOMETER_BUFF_LEN) anemometer_index = 0;

	return;
}

/*
 * Atomicly increment an index into e5850_buffer
 */
static inline void e5850_incr_bp(volatile unsigned long *index, int delta)
{
	unsigned long new = *index + delta;
	barrier ();  /* Don't optimize these two together */
	*index = (new >= (e5850_buffer + PAGE_SIZE)) ? e5850_buffer : new;
}

/*
 * Atomicly calculate e5850_buffer legend
 */
static inline  unsigned long e5850_buffer_len(void)
{
	long len = e5850_head - e5850_tail;
	
	if (len < 0) /* wrapped */
		len = e5850_buffer + PAGE_SIZE - e5850_tail;

	return (unsigned long) len;
}

int encoder_set_zero(int mask)
{
	DISABLE_INTERRUPT();
	outb_p( mask | PULSE_UP, e5850_base);
	mdelay(RELE_ACTIVE_TIME); 
	outb_p(        PULSE_UP, e5850_base);
	ENABLE_INTERRUPT();

	return 0;
}


int sidereal_clock_running(void)
{
	atomic_set( &test_irq,0);
	mdelay(TIME_OUT_SIDEREAL_CLK); /* whait for interrupt */
	return atomic_read(&test_irq);
}

void e5850_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int i, j;
	int v, val0 = 0, val1 = 0, val2 = 0;
	
	/* first triger */
	data_current &= ~ AAD_DESABLE;
	outb_p(PULSE_DOWN, e5850_base);
	/* main loop */
	for (i = e5850_nrb - 1, j = 9; i >=  0; i--, j--) {
		outb_p(PULSE_UP, e5850_base);
		/* for noise in transmission line (capacitance) */
		udelay(3); 
		v = inb_p(e5850_base + 1); 
		/* bit (v & MASK_X) >> OFFSET_X) inverted for 5850 */
		val0 |= (((v & MASK_0) >> OFFSET_0) ^ 1) << i;
		val1 |= (((v & MASK_1) >> OFFSET_1) ^ 1) << i;
		if (j >= 0 && j <9){
			val2 |= ((v & MASK_2) >> OFFSET_2) << j;
			if (j == 0)
				data_current |= AAD_DESABLE;
		}
		outb_p(PULSE_DOWN, e5850_base);
	}
	anemometer_push(val2);
	if (test_bit(0, &e5850_is_running)) {
		/* Write a 16 byte record. Assume PAGE_SIZE is a
                   multiple of 16 (E5850_DATA_LEN) */
		sprintf((char *)e5850_head,"%04x %04x %04x\n%c",
			val0, val1, 
			atomic_read(&anemometer_sum)>>ANEMOMETER_BUFF_SHIFT,0);
		e5850_incr_bp(&e5850_head, E5850_DATA_LEN);
		/* awake any reading process */
		wake_up_interruptible(&e5850_queue);
	}
	atomic_set( &encoder_0, val0);
	atomic_set( &encoder_1, val1);
	//anemometer_push(val2);
	atomic_set( &test_irq,1);
	outb_p(PULSE_UP, e5850_base);
}


/* only call this if test_bit(0, &e5850_fail_clock) is true */
void get_encoders_no_clock(void)
{
	unsigned long flags;

	if ( test_and_clear_bit(0, &e5850_is_running) ){
		printk(KERN_WARNING 
		       "Stoping circular buffer in fail mode.\n");

	}
	save_flags(flags); 
	cli();
	/* dummy read */
	e5850_interrupt(-1, NULL, NULL);
	restore_flags(flags);

	mdelay(1);

	save_flags(flags); 
	cli();
	/* real read */
	e5850_interrupt(-1, NULL, NULL);
	restore_flags(flags);


}

int put_on_fail_codition(void)
{
	if ( sidereal_clock_running() ) return -1;
	if ( test_bit(0, &e5850_fail_clock) ) return 1;

	DISABLE_INTERRUPT();
	if ( test_and_clear_bit(0, &e5850_is_running) ){
		printk(KERN_WARNING 
		       "Stoping circular buffer in fail mode.\n");
	}
	/* Clear circular buffer */
	e5850_head = e5850_tail = e5850_buffer;
	set_bit(0, &e5850_fail_clock);

	return 0;
}

int put_on_normal_codition(void)
{
	if ( !test_bit(0, &e5850_fail_clock) ) return 1;

	ENABLE_INTERRUPT();
	clear_bit(0, &e5850_fail_clock);
	if ( !sidereal_clock_running() ){
		printk(KERN_WARNING "Sidereal clock not running.\n");
		put_on_fail_codition();
		return -1;
	}

	return 0;
}


#ifdef USING_PROC_FS
static int e5850_atomic(char *buffer, char **start, off_t fpos, int length)
{
        char *temp;
	
        temp = buffer;
        temp += sprintf(temp, "Encoder0: 0x%04x\n", atomic_read(&encoder_0));
        temp += sprintf(temp, "Encoder1: 0x%04x\n", atomic_read(&encoder_1));
        temp += sprintf(temp, "Anemometer: 0x%02x\n", 
			atomic_read(&anemometer_sum) >> ANEMOMETER_BUFF_SHIFT);

        return temp-buffer;
}

static int e5850_info(char *buffer, char **start, off_t fpos, int length)
{
        char *temp;
	
        temp = buffer;
        temp += sprintf(temp, "Count:   %ld\n",e5850_buffer_len() >>4);
        temp += sprintf(temp, "Open:    %1d\n",test_bit(0, &e5850_is_open));
        temp += sprintf(temp, "Running: %1d\n",test_bit(0, &e5850_is_running));
        temp += sprintf(temp, "Sid CLK: %1d\n",sidereal_clock_running());
        temp += sprintf(temp, "FailMde: %1d\n",test_bit(0, &e5850_fail_clock));

        return temp-buffer;
}
#endif


/**
 *      e5850_open        -       handle open of e5850 encoder
 *      @inode: inode of device
 *      @file: file handle to device
 *
 *      Open the e5850 encoder device. Ensure only one person opens it
 *      at a time. Also start the circular buffer.
 */

static int e5850_open(struct inode *inode, struct file *file)
{
	if(test_and_set_bit(0, &e5850_is_open))
                return -EBUSY;
	set_bit(0, &e5850_is_running);

	return 0;
}

/**
 *      e5850_release     -       close 
 *      @inode: inode from VFS
 *      @file: file from VFS
 */

static int e5850_release(struct inode *inode, struct file *file)
{
	clear_bit(0, &e5850_is_running);
	clear_bit(0, &e5850_is_open);
	e5850_head = e5850_tail = e5850_buffer;
	
	return 0;
}

ssize_t e5850_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	int count0;

	/* This macro was added during 2.1 development to make life
           easier: */
	wait_event_interruptible(e5850_queue, e5850_head != e5850_tail);

	/* count0 is the number of readable data bytes */
	count0 = e5850_buffer_len();
	if (count0 < count) 
		count = count0;
	if (copy_to_user(buf, (char *)e5850_tail, count))
		return -EFAULT;
	e5850_incr_bp (&e5850_tail, count);
	return count;
}


/**
 *      e5850_ioctl       -       handle encoders ioctls
 *      @inode: inode of the device
 *      @file: file handle to the device
 *      @cmd: command
 *      @arg: argument pointer
 *
 *      Handle the encoders ioctls supported by the 5850 driver.
 **/

static int e5850_ioctl(struct inode *inode, struct file *file,
		       unsigned int cmd, unsigned long arg)
{
	int state;
	int enc;

	switch (cmd) {
	case E5850_GETBUFFLEN:
		return put_user((int) e5850_buffer_len(), (int *) arg);
	case E5850_CLEARBUFF:
		state = test_and_clear_bit(0, &e5850_is_running);
		e5850_head = e5850_tail = e5850_buffer;
		if(state) set_bit(0, &e5850_is_running);
		return 0;
	case E5850_START:
		set_bit(0, &e5850_is_running);
		return 0;
	case E5850_STOP:
		clear_bit(0, &e5850_is_running);
		return 0;
	case E5850_ISRUNNING:
		return test_bit(0, &e5850_is_running);
	case E5850_GETNVAL:
		if (get_user((int) enc,(int *) arg))
			return -EFAULT;
		if (test_bit(0, &e5850_fail_clock)) get_encoders_no_clock();
		switch ( enc ) {
		case 0:
			put_user((int) (atomic_read(&encoder_0)), (int *) arg);
			return 0;
		case 1:
			put_user((int) (atomic_read(&encoder_1)), (int *) arg);
			return 0;

		case 2:
			put_user((int) 
				 (atomic_read(&anemometer_sum) >> 
				  ANEMOMETER_BUFF_SHIFT), (int *) arg);
			return 0;

		default:
			return -1;
		}
	case E5850_SETCW:
		if (get_user((int) enc,(int *) arg))
			return -EFAULT;
		printk(KERN_INFO OUR_NAME": set cw on encoder %d\n", enc);
		switch ( enc ) {
		case 0:
			data_current &= ~ CCLOCKWISE_0;
			return 0;
		case 1:
			data_current &= ~ CCLOCKWISE_1;
			return 0;
		default:
			return -1;
		}
	case E5850_SETCCW:
		if (get_user((int) enc,(int *) arg))
			return -EFAULT;
		printk(KERN_INFO OUR_NAME": set ccw on encoder %d\n", enc);
		switch ( enc ) {
		case 0:
			data_current = CCLOCKWISE_0;
			return 0;
		case 1:
			data_current = CCLOCKWISE_1;
			return 0;
		default:
			return -1;
		}
	case E5850_SETZERO:
		if (get_user((int) enc,(int *) arg))
			return -EFAULT;
		printk(KERN_INFO OUR_NAME": set zero on encoder %d\n", enc);
		switch ( enc ) {
		case 0:
			return encoder_set_zero(SETZ_0);
		case 1:
			return encoder_set_zero(SETZ_1);
		default:
			return -1;
		}
	case E5850_SCLKRUNNING:
		return sidereal_clock_running();
	case E5850_FAILMODE:
		return put_on_fail_codition();
	case E5850_NORMALMODE:
		return put_on_normal_codition();
	case E5850_GETBITRES:
		return e5850_nrb;
        default:
                return -ENOIOCTLCMD;
		
        }
}


static struct file_operations e5850_fops = {
	.owner          = THIS_MODULE,
	.read           = e5850_read,
	.ioctl          = e5850_ioctl,
	.open           = e5850_open,
	.release        = e5850_release,
};


static int __init e5850_init(void)
{
	int result;
	int i;
	
	/**
	 * first, encoder5850 out the base/e5850_base ambiguity: we'd
	 * better use e5850_base in the code, for clarity, but allow
	 * setting just "base" at load time. Same for "irq" and
	 * "nrbits".
	 **/
	e5850_base = base;
	e5850_irq = irq;
	e5850_nrb = nrbits;

	/* Initial value for data port (base+0) */
	data_current = 0 | AAD_DESABLE;

	clear_bit(0, &e5850_is_running);
	clear_bit(0, &e5850_is_open);
	clear_bit(0, &e5850_fail_clock);

	result = check_region(e5850_base, 2);
	if (result) {
		printk(KERN_INFO OUR_NAME 
		       ": can't get I/O port address 0x%lx\n",
		       e5850_base);
		return result;
	}
	request_region(e5850_base, 2, OUR_NAME);
	
	result = register_chrdev(major, "e5850", &e5850_fops);
	if (result < 0) {
		printk(KERN_INFO OUR_NAME
		       ": can't get major number\n");
		release_region(e5850_base, 2);
		return result;
	}
	if (major == 0) major = result; /* dynamic */

	e5850_buffer = __get_free_pages(GFP_KERNEL,0); /* never fails */
	if (! e5850_buffer) {
		printk(KERN_ERR OUR_NAME
		       ": cannot get free pages for e5850_buffer\n");
	}
	e5850_head = e5850_tail = e5850_buffer;
	
	/**
	 * Set default interupt.
	 **/
	if (e5850_irq < 0) /* not yet specified: force the default on */
		switch (e5850_base) {
		case 0x378: e5850_irq = 7; break;
		case 0x278: e5850_irq = 2; break;
		case 0x3bc: e5850_irq = 5; break;
		}

	if (e5850_irq >= 0) {
		result = request_irq(e5850_irq, e5850_interrupt,
				     SA_INTERRUPT, OUR_NAME, NULL);
		if (result) {
			printk(KERN_INFO OUR_NAME 
			       ": can't get assigned irq %i\n",
			       e5850_irq);
			e5850_irq = -1;
    }
		else { /* actually enable it -- assume this *is* a
                          parallel port */
			ENABLE_INTERRUPT();
		}
	}

	atomic_set( &encoder_0, 0);
	atomic_set( &encoder_1, 0);
	atomic_set( &anemometer_sum, 0);
	for ( i = 0; i < ANEMOMETER_BUFF_LEN; i++) 
		anemometer_buff[i] = 0;

#ifdef USING_PROC_FS
	e5850_proc = proc_mkdir("e5850", 0);
	if (!e5850_proc) {
		printk (KERN_ERR "cannot init /proc/e5850\n");
		return -ENOMEM;
	}
	create_proc_info_entry("e5850/info", 0, NULL, e5850_info);
	create_proc_info_entry("e5850/current_val", 0, NULL, e5850_atomic);
#endif
	
	/* Up Pulse Line */
	outb_p(PULSE_UP, e5850_base);   
	
	printk(KERN_INFO "Absolute encoder 5850 v(" 
	       E5850_VERSION ") (%d bits) initialising at io 0x%lx irq %d.\n",
	       e5850_nrb, e5850_base, e5850_irq);

	if ( ! sidereal_clock_running() ) {
		printk(KERN_WARNING 
		       "Sidereal clock not detected at startup.\n");
	}
	return 0;
}


static void __exit e5850_exit(void)
{
	if (e5850_irq >= 0) {
		DISABLE_INTERRUPT();   /* disable the interrupt */
		free_irq(e5850_irq, NULL);
	}
	/* Up Pulse Line */
	outb_p(PULSE_UP, e5850_base);
	unregister_chrdev(major, "e5850");
	release_region(e5850_base, 2);
#ifdef USING_PROC_FS
	remove_proc_entry("e5850/info",NULL);
	remove_proc_entry("e5850/current_val",NULL);
	remove_proc_entry ("e5850", 0);
#endif
	if (e5850_buffer) free_page(e5850_buffer);

	printk(KERN_INFO "Absolute encoder 5850 v(" 
	       E5850_VERSION ") down.\n");
	return;
}

module_init(e5850_init);
module_exit(e5850_exit);

MODULE_AUTHOR("Federico Bareilles <fede@iar.unlp.edu.ar>");
MODULE_DESCRIPTION("5850 Absolute encoder");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("integrated PC-style parallel port");

EXPORT_NO_SYMBOLS;
